local E_MODEL_FOG = smlua_model_util_get_id("fog_geo")

local STATE_NORMAL = 0
local STATE_SAND = 1
local STATE_BLACK = 2
local STATE_GREEN = 3
local STATE_PURPLE = 4
local STATE_HAUNTED = 5
local STATE_FIRE = 6

local skyboxInfo = {
    [BACKGROUND_OCEAN_SKY]       = {anim = STATE_BLACK , color = {r = 150, g = 150, b = 150}}, -- {r = 255, g = 162, b = 190}},
    [BACKGROUND_SNOW_MOUNTAINS]  = {anim = STATE_BLACK , color = {r = 150, g = 150, b = 150}}, -- {r = 255, g = 162, b = 190}},
    [BACKGROUND_ABOVE_CLOUDS]    = {anim = STATE_BLACK , color = {r = 150, g = 150, b = 150}}, -- {r = 255, g = 162, b = 190}},
    [BACKGROUND_BELOW_CLOUDS]    = {anim = STATE_BLACK , color = {r = 150, g = 150, b = 150}}, -- {r = 255, g = 162, b = 190}},
    [BACKGROUND_UNDERWATER_CITY] = {anim = STATE_BLACK , color = {r = 150, g = 150, b = 150}}, -- {r = 255, g = 162, b = 190}},
    [BACKGROUND_FLAMING_SKY]     = {anim = STATE_BLACK , color = {r = 150, g = 150, b = 150}}, -- {r = 255, g = 162, b = 190}},
    [BACKGROUND_GREEN_SKY]       = {anim = STATE_BLACK , color = {r = 150, g = 150, b = 150}}, -- {r = 255, g = 162, b = 190}},
    [BACKGROUND_HAUNTED]         = {anim = STATE_BLACK , color = {r = 150, g = 150, b = 150}}, -- {r = 255, g = 162, b = 190}},
    [BACKGROUND_DESERT]          = {anim = STATE_BLACK , color = {r = 150, g = 150, b = 150}}, -- {r = 255, g = 162, b = 190}},
    [BACKGROUND_PURPLE_SKY]      = {anim = STATE_BLACK , color = {r = 150, g = 150, b = 150}}, -- {r = 255, g = 162, b = 190}},
    [BACKGROUND_CUSTOM]          = {anim = STATE_BLACK , color = {r = 150, g = 150, b = 150}}, -- {r = 255, g = 162, b = 190}},
}

---@param o Object
local function fog_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    o.header.gfx.skipInViewCheck = true
    o.oFaceAnglePitch = 0
    o.oFaceAngleRoll = 0
    o.oOpacity = 75
    obj_scale(o, 3.5)
    set_override_far(1000000)
end

local function get_color_from_skybox(skybox, m)
    local r, g, b = 100, 100, 100

    if skybox >= 0 then
        r, g, b = skyboxInfo[skybox].color.r, skyboxInfo[skybox].color.g, skyboxInfo[skybox].color.b
    end

    -- check skybox to determine the color of the fog
    if skybox == BACKGROUND_CUSTOM then
        -- custom skybox, use terrain type instead
        if m.area.terrainType == TERRAIN_SAND then
            r, g, b = skyboxInfo[BACKGROUND_DESERT].color.r, skyboxInfo[BACKGROUND_DESERT].color.g, skyboxInfo[BACKGROUND_DESERT].color.b
        elseif m.area.terrainType == TERRAIN_SPOOKY then
            r, g, b = skyboxInfo[BACKGROUND_HAUNTED].color.r, skyboxInfo[BACKGROUND_HAUNTED].color.g, skyboxInfo[BACKGROUND_HAUNTED].color.b
        else
            r, g, b = 100, 100, 100
        end
    end
    return r, g, b
end


---@param o Object
local function fog_loop(o)

    local r, g, b = get_color_from_skybox(get_skybox(), gMarioStates[0])

    set_lighting_color(0, r)
    set_lighting_color(1, g)
    set_lighting_color(2, b)
    set_vertex_color(0, r)
    set_vertex_color(1, g)
    set_vertex_color(2, b)
    set_fog_color(0, clamp(r - (r * 0.75), 0, 255))
    set_fog_color(1, clamp(g - (g * 0.75), 0, 255))
    set_fog_color(2, clamp(b - (b * 0.75), 0, 255))

    if not gGlobalSyncTable.isSwapped then
        obj_mark_for_deletion(o)

        set_lighting_color(0, 255)
        set_lighting_color(1, 255)
        set_lighting_color(2, 255)
        set_vertex_color(0, 255)
        set_vertex_color(1, 255)
        set_vertex_color(2, 255)
        set_fog_color(0, 255)
        set_fog_color(1, 255)
        set_fog_color(2, 255)
    end
end

local function update()

    local m = gMarioStates[0]

    local fog = obj_get_first_with_behavior_id(id_bhvFog)

    if not fog and gGlobalSyncTable.isSwapped then
        spawn_non_sync_object(id_bhvFog, E_MODEL_FOG, 0, 0, 0, nil)
    elseif fog then
        --* in the update hook so it also follows the camera in the pause menu and other things where a behavior wouldn't normally be able to run
        fog.oPosX, fog.oPosY, fog.oPosZ = gLakituState.pos.x, gLakituState.pos.y, gLakituState.pos.z

        fog.oBehParams2ndByte = get_skybox() --* in case mods change the skybox this runs constantly

        if skyboxInfo[fog.oBehParams2ndByte] then
            fog.oAnimState = skyboxInfo[fog.oBehParams2ndByte].anim
        else
            fog.oAnimState = STATE_BLACK
        end
    end
end

id_bhvFog = hook_behavior(nil, OBJ_LIST_DEFAULT, false, fog_init, fog_loop)
hook_event(HOOK_UPDATE, update)