local currentMenu = nil
local currentIndex = 1
local scrollIndices = {}
local scrollOffset = 0
local lockedIndex = nil

local MAX_VISIBLE_OPTIONS = 6
local textHeight = 6

local function set_menu(menu)
    currentMenu = {}
    currentIndex = 1
    scrollIndices = {}
    scrollOffset = 0
    lockedIndex = nil
    
    local isHost = network_is_server()

    for i = 1, #menu do
        if not menu[i].hostOnly or isHost then
            table.insert(currentMenu, menu[i])
            scrollIndices[#currentMenu] = 1
        end
    end
end

local function toggle_option(option)
    if option.enableable ~= nil then
        option.enableable = not option.enableable
    end
end

local function update_scroll_offset()
    if currentIndex > scrollOffset + MAX_VISIBLE_OPTIONS then
        scrollOffset = currentIndex - MAX_VISIBLE_OPTIONS
    elseif currentIndex <= scrollOffset then
        scrollOffset = currentIndex - 1
    end
end

sMainMenuOptions = {
    {
        name = "Spawn Entity",
        enableable = true,
        func = function(option)
            gGlobalSyncTable.spawnEntity = not gGlobalSyncTable.spawnEntity
        end
    },
    {
        name = "Remove Stars On Death",
        enableable = true,
        func = function(option)
            gGlobalSyncTable.removeStarsOnDeath = not gGlobalSyncTable.removeStarsOnDeath
        end
    },
    {
        name = "Coins Needed To Swap",
        value = INITIAL_MAX_POINTS,
        valueText = " Coins",
        valueTextMode = "append",
        min = 10,
        max = 100,
        step = 10,
        func = function(option)
            INITIAL_MAX_POINTS = option.value
        end
    },
    {
        name = "Entity Speed",
        value = gGlobalSyncTable.entitySpeed,
        valueText = " Units per frame",
        valueTextMode = "append",
        min = 5,
        max = 100,
        step = 5,
        func = function(option)
            gGlobalSyncTable.entitySpeed = option.value
        end
    },
    { name = "BACK", func = function()
        mainMenu = false
        currentIndex = 1
        scrollIndices = {}
        scrollOffset = 0
        lockedIndex = nil
        for i = 1, #currentMenu do
            scrollIndices[i] = 1
        end
    end
    },
}

local function render_menu()
    local sWidth = djui_hud_get_screen_width()
    local sHeight = djui_hud_get_screen_height()
    local lineHeight = 20
    local scale = 1
    local visibleOptions = math.min(#currentMenu - scrollOffset, MAX_VISIBLE_OPTIONS)
    local y = 40

    for i = scrollOffset + 1, scrollOffset + visibleOptions do
        local option = currentMenu[i]
        if not option then break end

        local text = type(option.name) == "function" and option.name() or option.name

        if option.scrollOptions then
            local scrollIndex = scrollIndices[i]
            local scrollOption = option.scrollOptions[scrollIndex]
            text = text .. ": " .. scrollOption.name
        elseif option.value then
            if option.valueText then
                if option.valueTextMode == "replace" then
                    text = text .. ": " .. tostring(option.valueText(option))
                else
                    text = text .. ": " .. tostring(option.value) .. tostring(option.valueText)
                end
            else
                text = text .. ": " .. tostring(option.value)
            end
        end

        if i == currentIndex then
            if option.waitingForInput then
                text = ">> " .. text .. " <<"
            elseif lockedIndex == i then
                text = ">> " .. text .. " <<"
            else
                text = "> " .. text
            end
        end

        djui_hud_set_font(FONT_TINY)
        local textWidth = djui_hud_measure_text(text)

        local x = (sWidth * 0.5) - (textWidth * 0.5)

        local padding = 5
        local rectWidth = textWidth + padding * 2
        local rectHeight = textHeight + padding * 2

        if option.enableable then
            djui_hud_set_color(34, 200, 34, 128)
        elseif option.enableable == false then
            djui_hud_set_color(255, 0, 0, 128)
        else
            djui_hud_set_color(0, 0, 0, 128)
        end

        djui_hud_render_rect(x - padding, y, rectWidth, rectHeight)

        djui_hud_set_font(FONT_TINY)
        djui_hud_set_color(255, 255, 255, 255)
        djui_hud_print_text(text, x, y, scale)

        y = y + lineHeight
    end

    if scrollOffset + MAX_VISIBLE_OPTIONS < #currentMenu then
        local dotsX = sWidth * 0.5 - 1.5
        local dotsY = y
        local padding = 5
        local rectWidth = 3 + padding * 2
        local rectHeight = textHeight + padding * 2

        djui_hud_set_color(0, 0, 0, 128)
        djui_hud_render_rect(dotsX - 1, y, rectWidth, rectHeight)
        djui_hud_set_color(255, 255, 255, 255)
        djui_hud_print_text("...", dotsX, dotsY, 1)
    end
end

local function handle_input()
    local m = gMarioStates[0]
    m.freeze = 15

    local func = currentMenu[currentIndex].receivedInput

    if currentMenu[currentIndex].waitingForInput then
        if m.controller.buttonPressed ~= 0 then
            currentMenu[currentIndex].waitingForInput = false
            play_sound(SOUND_ACTION_READ_SIGN, gGlobalSoundSource)
            if func then
                func(currentMenu[currentIndex])
            end
        end
        return
    end

    if lockedIndex then
        local option = currentMenu[lockedIndex]
        if m.controller.buttonPressed & A_BUTTON ~= 0 then
            play_sound(SOUND_ACTION_READ_SIGN, gGlobalSoundSource)
            lockedIndex = nil
            if option.func then
                option.func(option)
                if option.enableable ~= nil then
                    toggle_option(option)
                end
            end
        elseif m.controller.buttonPressed & L_JPAD ~= 0 then
            play_sound(SOUND_MENU_CHANGE_SELECT, gGlobalSoundSource)
            if option.scrollOptions then
                local scrollIndex = scrollIndices[lockedIndex] - 1
                if scrollIndex < 1 then scrollIndex = #option.scrollOptions end
                scrollIndices[lockedIndex] = scrollIndex
            elseif option.value then
                option.value = math.max(option.min, option.value - option.step)
            end
        elseif m.controller.buttonPressed & R_JPAD ~= 0 then
            play_sound(SOUND_MENU_CHANGE_SELECT, gGlobalSoundSource)
            if option.scrollOptions then
                local scrollIndex = scrollIndices[lockedIndex] + 1
                if scrollIndex > #option.scrollOptions then scrollIndex = 1 end
                scrollIndices[lockedIndex] = scrollIndex
            elseif option.value then
                option.value = math.min(option.max, option.value + option.step)
            end
        end
    else
        if m.controller.buttonPressed & U_JPAD ~= 0 then
            play_sound(SOUND_MENU_CHANGE_SELECT, gGlobalSoundSource)
            currentIndex = currentIndex - 1
            if currentIndex < 1 then currentIndex = #currentMenu end
            update_scroll_offset()
        elseif m.controller.buttonPressed & D_JPAD ~= 0 then
            play_sound(SOUND_MENU_CHANGE_SELECT, gGlobalSoundSource)
            currentIndex = currentIndex + 1
            if currentIndex > #currentMenu then currentIndex = 1 end
            update_scroll_offset()
        elseif m.controller.buttonPressed & A_BUTTON ~= 0 or m.controller.buttonPressed & B_BUTTON ~= 0 then
            play_sound(SOUND_ACTION_READ_SIGN, gGlobalSoundSource)
            local option = currentMenu[currentIndex]
            if option.func and (not option.value and not option.scrollOptions) then
                if option.enableable ~= nil then
                    toggle_option(option)
                end
                option.func(option)
            elseif option.value or option.scrollOptions then
                lockedIndex = currentIndex
            end
        end
    end
end

local function hud_render()
    djui_hud_set_resolution(RESOLUTION_N64)
    djui_hud_set_font(FONT_TINY)

    if not mainMenu or not currentMenu then return end
    handle_input()
    render_menu()
end

set_menu(sMainMenuOptions)

hook_event(HOOK_ON_HUD_RENDER, hud_render)

local function open_menu()
    mainMenu = not mainMenu
    return true
end

if network_is_server() then
    hook_chat_command("aocm", "Opens the menu", open_menu)
end